from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="product_attribute_source_")

    tablename: str
    region_name: str
    name_language_indexname: str
